/*

    Copyright 2004 Bastiaan Veelo

    This file is part of dcouple.

    Dcouple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    You are free to negociate a different license with the copyright holder.
    
    Dcouple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dcouple; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Some concepts in this file were inspired by Andy Friesen's listener.d
    (http://andy.tadan.us/d/listener.d).
    
*/

module dcouple.signal;

private import dcouple.slot;
private import dcouple.signalslot;

/+
//! An interface for the management of signals.
/*! Provides an interface that must be implemented by classes that own \link Signal Signals \endlink in a managed way. */
interface SignalManager
{
	/*! Deriving classes must implement this function to keep a list of 
	\link Signal Signals \endlink that it owns, and disconnect() them in the 
	destructor (so that connected \link Slot Slots \endlink remove their 
	reference). Also, deriving classes must provide and use a method (say, 
	.discard()) to disconnect all signals when the object is no longer used, 
	so that it may be garbage collected.
	*/
	void register(GenericSignal);
}
+/

//! An interface shared by all Signals
//interface GenericSignal
class GenericSignal
{
	//! Disconnect all. 
	/*! Breaks all connections from this Signal. */
	//void disconnect();
	void disconnect() {}
	SignalSlotManager owner() { return null; }
	
	void genericConnect(GenericSlot s) {}
	void genericDisconnect(GenericSlot s) {}
	
	uint numberOfArguments() {return 0;}
	char[][] arguments() {return DUMMYarguments;}
private:
	char[][] DUMMYarguments; // Remove this when we can be an interface.
}


/*! The generic part of \link Signal Signals. \endlink */
private template SignalGenericCore() {
	//! Constructor.
	/*! The Signal \link SignalManager::register(Signal) registers \endlink 
	itself with its owner. 
	*/
	//this(SignalManager owner)
	this(SignalSlotManager owner)
	{
		_owner = owner;
		owner.register(this);
		initArguments();
	}
	
	//! Destructor.
	/*! Disconnects all connected \link Slot Slots \endlink, and \link 
	SignalManager::deregister(Signal) de-registers \endlink itself with its 
	owner. 
	*/
	~this()
	{
		disconnect();
		/* Strictly spoken, destructors are not allowed to reference 
		other objects, because the order in which the GC calls 
		destructors is not defined, so those objects may not exist 
		anymore. The disconnect() function above breaks this rule 
		indirectly. But this particular case is OK, because we only 
		consider references to CompatibleSlots that *are still alive*. 
		Slots that are already destructed have removed our reference to 
		them, just like we have removed references to us in other slots 
		in disconnect(). */
	}
	
	/*! Establishes a connection between this Signal and Slot \a s, and vice 
	versa (i.e., registers this Signal within Slot \a s).
	*/
	void connect(CompatibleSlot s)
	{
		if(s in _slots) {} else {
			_slots[s] = s;
			s.connect(this);
		}
	}
	
	void genericConnect(GenericSlot gs)
	{
		CompatibleSlot s = cast(CompatibleSlot) gs;
		if( s ) {
			connect(s);
		} else {
			printf("dcouple WARNING: %.*s.%.*s!(", 
				_owner.classinfo.name, this.classinfo.name);
			if(numberOfArguments()>0) printf("%.*s",arguments()[0]);
			for(int i=1; i<numberOfArguments();i++)
				printf(",%.*s",arguments()[i]);
			printf(") is incompatible with %.*s.%.*s!(", 
				gs.owner().classinfo.name,  gs.classinfo.name);
			if(gs.numberOfArguments()>0)
				printf("%.*s",gs.arguments()[0]);
			for(int i=1; i<gs.numberOfArguments();i++)
				printf(",%.*s",gs.arguments()[i]);
			printf("); not connected. (Sorry, no line number...)\n");
		}
	}
	
	/*! Breaks the connection between this Signal and Slot \a s (if it 
	exists), and vice versa (i.e., de-registers this Signal from Slot \a s).
	*/
	void disconnect(CompatibleSlot s)
	{
		if(s in _slots) {
			delete _slots[s];
			s.disconnect(this);
		}
	}
	
	void genericDisconnect(GenericSlot gs)
	{
		CompatibleSlot s = cast(CompatibleSlot) gs;
		if( s ) {
			disconnect(s);
		} else {
			printf("dcouple WARNING: %.*s.%.*s!(", 
				_owner.classinfo.name, this.classinfo.name);
			if(numberOfArguments()>0) printf("%.*s",arguments()[0]);
			for(int i=1; i<numberOfArguments();i++)
				printf(",%.*s",arguments()[i]);
			printf(") is incompatible with %.*s.%.*s!(", 
				gs.owner().classinfo.name,  gs.classinfo.name);
			if(gs.numberOfArguments()>0)
				printf("%.*s",gs.arguments()[0]);
			for(int i=1; i<gs.numberOfArguments();i++)
				printf(",%.*s",gs.arguments()[i]);
			printf("); not disconnected. (Sorry, no line " 
				"number...)\n");
		}
	}
	
	//! Disconnect all. 
	/*! Breaks all connections from this Signal. */
	void disconnect()
	{
		foreach(CompatibleSlot s; _slots) disconnect(s);
	}

	/*! \returns the number of connected \link Slot Slots. \endlink */
	int count() {
        	return _slots.length;
	}
	
	uint numberOfArguments() 
	{
		return _arguments.length;
	}
	
	char[][] arguments() {return _arguments;}
	
	SignalSlotManager owner() { return _owner; }
private:
	//! This Signal is managed by _owner.
	//SignalManager _owner;
	SignalSlotManager _owner;
protected:
	//! Collection of connected \link Slot Slots. \endlink
	CompatibleSlot[CompatibleSlot] _slots;	// Crashes.
	char[][] _arguments;
}

//! A signal for the connection to slots.
/*! A Signal should be owned by a \link SignalManager managing \endlink class,
and can be \link connect() connected \endlink to one or more \link Slot Slots, \endlink possibly in other classes. These \link Slot Slots \endlink are notified
when the Signal is \link emit() emitted. \endlink
*/
class Signal() : GenericSignal
{
	alias Slot!() CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit()
	{
		foreach( CompatibleSlot s; _slots ) s();
	}
	
	void initArguments() {}
}

class Signal(T1) : GenericSignal
{
	alias Slot!(T1) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1)
	{
		foreach( CompatibleSlot s; _slots ) s(t1);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
	}
}

class Signal(T1,T2) : GenericSignal
{
	alias Slot!(T1,T2) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
	}
}

class Signal(T1,T2,T3) : GenericSignal
{
	alias Slot!(T1,T2,T3) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2, T3 t3)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2, t3);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
	}
}

class Signal(T1,T2,T3,T4) : GenericSignal
{
	alias Slot!(T1,T2,T3,T4) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2, T3 t3, T4 t4)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2, t3, t4);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
	}
}

class Signal(T1,T2,T3,T4,T5) : GenericSignal
{
	alias Slot!(T1,T2,T3,T4,T5) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2, t3, t4, t5);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
	}
}

class Signal(T1,T2,T3,T4,T5,T6) : GenericSignal
{
	alias Slot!(T1,T2,T3,T4,T5,T6) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2, t3, t4, t5, t6);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
		_arguments ~= typeid(T6).toString();
	}
}

class Signal(T1,T2,T3,T4,T5,T6,T7) : GenericSignal
{
	alias Slot!(T1,T2,T3,T4,T5,T6,T7) CompatibleSlot;
	mixin SignalGenericCore;
	
	/*! Causes all connected \link Slot Slots \endlink to be notified. */
	void emit(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7)
	{
		foreach( CompatibleSlot s; _slots ) s(t1, t2, t3, t4, t5, t6, t7);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
		_arguments ~= typeid(T6).toString();
		_arguments ~= typeid(T7).toString();
	}
}
