/*

    Copyright 2004 Bastiaan Veelo

    This file is part of dcouple.

    Dcouple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    You are free to negociate a different license with the copyright holder.
    
    Dcouple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dcouple; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

module dcouple.slot;

private import dcouple.signal;
private import dcouple.signalslot;

/+
//! An interface for the management of slots.
/*! Provides an interface that must be implemented by classes that own \link Slot Slots \endlink in a managed way. */
interface SlotManager
{
	/*! Deriving classes must implement this function to keep a list of \link 
	Slot Slots \endlink that it owns, and disconnect() them in the destructor 
	(so that connected \link Signal Signals \endlink remove their reference). 
	Also, deriving classes must provide and use a method (say, .discard()) to 
	disconnect all slots when the object is no longer used, so that it may be 
	garbage collected.
	*/
	void register(GenericSlot);
}
+/

//! An interface shared by all Slots
//interface GenericSlot
class GenericSlot
{
	//! Disconnect all. 
	/*! Breaks all connections from this Slot. */
//	void disconnect();
	void disconnect() {}
	SignalSlotManager owner() { return null; }
	
	void genericConnect(GenericSignal s) {}
	void genericDisconnect(GenericSignal s) {}
	
	uint numberOfArguments() {return 0;}
	char[][] arguments() {return DUMMYarguments;}
private:
	char[][] DUMMYarguments; // Remove this when we can be an interface.
}

/*! The generic part of \link Slot Slots. \endlink */
private template SlotGenericCore() {
	//! Constructor.
	/*! The Slot \link SlotManager::register(Slot) registers \endlink 
	itself with its \a owner. \a callBack must be a member function of \a 
	owner, and is the method that is called when a signal is received.
	*/
//	this(SlotManager owner, CallBack callBack)
	this(SignalSlotManager owner, CallBack callBack)
	{
		_owner = owner;
		_callBack = callBack;
		owner.register(this);
		initArguments();
	}
	
	//! Destructor.
	/*! Disconnects all connected \link Signal Signals \endlink, and \link 
	SlotManager::deregister(Slot) de-registers \endlink itself with its 
	owner. 
	*/
	~this()
	{
		disconnect();
		/* Strictly spoken, destructors are not allowed to reference 
		other objects, because the order in which the GC calls 
		destructors is not defined, so those objects may not exist 
		anymore. The disconnect() function above breaks this rule 
		indirectly. But this particular case is OK, because we only 
		consider references to CompatibleSignals that *are still alive*. 
		Signals that are already destructed have removed our reference to 
		them, just like we have removed references to us in other signals 
		in disconnect(). */
	}
	
	/*! Establishes a connection between this Slot and Signal \a s, and vice 
	versa (i.e., registers this Slot within Signal \a s).
	*/
	void connect(CompatibleSignal s)
	{
		if(s in _signals) {} else {
			_signals[s] = s;
			s.connect(this);
		}
	}
	
	void genericConnect(GenericSignal gs)
	{
		CompatibleSignal s = cast(CompatibleSignal) gs;
		if( s ) {
			connect(s);
		} else {
			printf("dcouple WARNING: %.*s.%.*s!(", 
				_owner.classinfo.name, this.classinfo.name);
			if(numberOfArguments()>0) printf("%.*s",arguments()[0]);
			for(int i=1; i<numberOfArguments();i++)
				printf(",%.*s",arguments()[i]);
			printf(") is incompatible with %.*s.%.*s!(", 
				gs.owner().classinfo.name,  gs.classinfo.name);
			if(gs.numberOfArguments()>0)
				printf("%.*s",gs.arguments()[0]);
			for(int i=1; i<gs.numberOfArguments();i++)
				printf(",%.*s",gs.arguments()[i]);
			printf("); not connected. (Sorry, no line number...)\n");
		}
	}
	
	/*! Breaks the connection between this Slot and Signal \a s (if it 
	exists), and vice versa (i.e., de-registers this Slot from Signal \a s).
	*/
	void disconnect(CompatibleSignal s)
	{
		if(s in _signals) {
			delete _signals[s];
			s.disconnect(this);
		}
	}
	
	void genericDisconnect(GenericSignal gs)
	{
		CompatibleSignal s = cast(CompatibleSignal) gs;
		if( s ) {
			disconnect(s);
		} else {
			printf("dcouple WARNING: %.*s.%.*s!(", 
				_owner.classinfo.name, this.classinfo.name);
			if(numberOfArguments()>0) printf("%.*s",arguments()[0]);
			for(int i=1; i<numberOfArguments();i++)
				printf(",%.*s",arguments()[i]);
			printf(") is incompatible with %.*s.%.*s!(", 
				gs.owner().classinfo.name,  gs.classinfo.name);
			if(gs.numberOfArguments()>0)
				printf("%.*s",gs.arguments()[0]);
			for(int i=1; i<gs.numberOfArguments();i++)
				printf(",%.*s",gs.arguments()[i]);
			printf("); not disconnected. (Sorry, no line number...)\n");
		}
	}
	
	//! Disconnect all. 
	/*! Breaks all connections from this Slot. */
	void disconnect()
	{
		foreach(CompatibleSignal s; _signals) {
			disconnect(s);
		}
	}

	/*! \returns the number of connected \link Signal Signals. \endlink */
	int count() {
        	return _signals.length;
	}
	
	uint numberOfArguments() 
	{
		return _arguments.length;
	}
	
	char[][] arguments() {return _arguments;}

	SignalSlotManager owner() { return _owner; }
protected:
	//! This Slot is managed by _owner.
	//SlotManager _owner;
	SignalSlotManager _owner;
	//! "delegate", or pointer-to-member-function.
	CallBack _callBack;
	//! Collection of connected \link Signal Signals. \endlink
	CompatibleSignal[CompatibleSignal] _signals;
	char[][] _arguments;
}



//! A slot to which signals can be connected.
/*! A Slot should be owned by a \link SlotManager managing \endlink class,
and can be \link connect() connected \endlink to one or more \link Signal Signals, \endlink possibly in other classes. This Slot is then notified when a connected Signal is \link Signal::emit() emitted, \endlink and the corresponding call-back function ("delegate") called.
*/
class Slot() : GenericSlot
{
	alias void delegate() CallBack;
	alias Signal!() CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall()
	{
		_callBack();
	}
	void initArguments() {}
}

class Slot(T1) : GenericSlot
{
	alias void delegate(T1) CallBack;
	alias Signal!(T1) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1)
	{
		_callBack(t1);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
	}
}

class Slot(T1, T2) : GenericSlot
{
	alias void delegate(T1,T2) CallBack;
	alias Signal!(T1,T2) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2)
	{
		_callBack(t1, t2);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
	}
}

class Slot(T1,T2,T3) : GenericSlot
{
	alias void delegate(T1,T2,T3) CallBack;
	alias Signal!(T1,T2,T3) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2, T3 t3)
	{
		_callBack(t1, t2, t3);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
	}
}

class Slot(T1,T2,T3,T4) : GenericSlot
{
	alias void delegate(T1,T2,T3,T4) CallBack;
	alias Signal!(T1,T2,T3,T4) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2, T3 t3, T4 t4)
	{
		_callBack(t1, t2, t3, t4);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
	}
}

class Slot(T1,T2,T3,T4,T5) : GenericSlot
{
	alias void delegate(T1,T2,T3,T4,T5) CallBack;
	alias Signal!(T1,T2,T3,T4,T5) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5)
	{
		_callBack(t1, t2, t3, t4, t5);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
	}
}

class Slot(T1,T2,T3,T4,T5,T6) : GenericSlot
{
	alias void delegate(T1,T2,T3,T4,T5,T6) CallBack;
	alias Signal!(T1,T2,T3,T4,T5,T6) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6)
	{
		_callBack(t1, t2, t3, t4, t5, t6);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
		_arguments ~= typeid(T6).toString();
	}
}

class Slot(T1,T2,T3,T4,T5,T6,T7) : GenericSlot
{
	alias void delegate(T1,T2,T3,T4,T5,T6,T7) CallBack;
	alias Signal!(T1,T2,T3,T4,T5,T6,T7) CompatibleSignal;
	mixin SlotGenericCore;

	/*! Calls the registered callBack member function. Allows the 
	Slot to be called like a function.
	*/
	void opCall(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7)
	{
		_callBack(t1, t2, t3, t4, t5, t6, t7);
	}
	
	void initArguments()
	{
		_arguments ~= typeid(T1).toString();
		_arguments ~= typeid(T2).toString();
		_arguments ~= typeid(T3).toString();
		_arguments ~= typeid(T4).toString();
		_arguments ~= typeid(T5).toString();
		_arguments ~= typeid(T6).toString();
		_arguments ~= typeid(T7).toString();
	}
}
